# Copyright 2016 Stephan Rave. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
#   disclaimer in the documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from functools import partial

import numpy as np

from pymor.analyticalproblems.burgers import BurgersProblem, Burgers2DProblem
from pymor.discretizers.advection import discretize_nonlinear_instationary_advection_fv
from pymor.domaindiscretizers.default import discretize_domain_default
from pymor.grids.rect import RectGrid

from discretization import freeze


# 1D example
p = BurgersProblem(circle=True)
d, d_data = discretize_nonlinear_instationary_advection_fv(
    p,
    diameter=np.sqrt(2) / 480,
    nt=3200
)
d_f = freeze(d, d_data)

U = d.solve(2)
U_frozen = d_f.solve(2)
U_unfrozen = U_frozen.empty(reserve=len(U_frozen))
for n in range(len(U_frozen)):
    u = U_frozen.copy(n)
    ug = d_f.group_action.apply(U_frozen.group[n], u)
    U_unfrozen.append(ug)
d.visualize((U, U_unfrozen), legend=('not frozen', 'frozen'))


# 2D example
p = Burgers2DProblem()
d, d_data = discretize_nonlinear_instationary_advection_fv(
    p,
    diameter=np.sqrt(2) / 60,
    nt=100,
    num_flux='lax_friedrichs',
    lxf_lambda=0.5,
    domain_discretizer=partial(discretize_domain_default, grid_type=RectGrid)
)
d_f = freeze(d, d_data)

U = d.solve(2)
U_frozen = d_f.solve(2)
U_unfrozen = U_frozen.empty(reserve=len(U_frozen))
for n in range(len(U_frozen)):
    u = U_frozen.copy(n)
    ug = d_f.group_action.apply(U_frozen.group[n], u)
    U_unfrozen.append(ug)
d.visualize((U, U_unfrozen), legend=('not frozen', 'frozen'))
