# Copyright 2016 Stephan Rave. All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
#   disclaimer in the documentation and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# Settings
DIAMETER      = 1. / 120
NT            = 200
VISUALIZE_MU  = None
RB_SAMPLES    = 10
CB_SAMPLES    = 10
RB_SIZE       = 100
CB_SIZE       = 200
CB_RB_RATIOS  = [1.8, 1.9]
TEST_SET_SIZE = 100
TEST_DIMS     = 20


from functools import partial
import sys

import numpy as np

from pymor.analyticalproblems.burgers import Burgers2DProblem
from pymor.domaindiscretizers.default import discretize_domain_default
from pymor.grids.rect import RectGrid
from pymor.discretizers.advection import discretize_nonlinear_instationary_advection_fv
from pymor.algorithms.ei import ei_greedy
from pymor.algorithms.greedy import greedy
from pymor.algorithms.basisextension import pod_basis_extension
from pymor.operators.ei import EmpiricalInterpolatedOperator
from pymor.reductors.basic import GenericRBReconstructor

from group import OrthogonalityPhaseCondition
from discretization import freeze, FrozenDiscretization


def reduce_frozen(d, rb, collateral_basis, dofs, N=None, M=None, extends=None):
    collateral_basis_M = collateral_basis if M is None else collateral_basis.copy(ind=list(range(M)))
    dofs_M = dofs if M is None else dofs[:M]
    op = EmpiricalInterpolatedOperator(d.operator, dofs_M, collateral_basis_M, triangular=True)
    frozen_op = EmpiricalInterpolatedOperator(d.frozen_operator, dofs_M, collateral_basis_M, triangular=True)
    phase_condition = OrthogonalityPhaseCondition(d.group_action, op)

    if rb is None:
        rb = d.solution_space.empty()
    rb_N = rb if N is None else rb.copy(ind=list(range(N)))
    op = op.projected(source_basis=rb_N, range_basis=rb_N)
    frozen_op = frozen_op.projected(source_basis=rb_N, range_basis=rb_N)
    phase_condition = phase_condition.projected(rb_N)
    initial_data = d.initial_data.projected(source_basis=None, range_basis=rb_N)

    rd = d.with_(operator=op, frozen_operator=frozen_op, phase_condition=phase_condition,
                 initial_data=initial_data)
    rd.disable_logging()
    rc = GenericRBReconstructor(rb_N)
    return rd, rc, {}


def reduce_nonfrozen(d, rb, collateral_basis, dofs, N=None, M=None, extends=None):
    collateral_basis_M = collateral_basis if M is None else collateral_basis.copy(ind=list(range(M)))
    dofs_M = dofs if M is None else dofs[:M]
    op = EmpiricalInterpolatedOperator(d.operator, dofs_M, collateral_basis_M, triangular=True)
    if rb is None:
        rb = d.solution_space.empty()
    rb_N = rb if N is None else rb.copy(ind=list(range(N)))
    op = op.projected(source_basis=rb_N, range_basis=rb_N)
    initial_data = d.initial_data.projected(source_basis=None, range_basis=rb_N)
    rd = d.with_(operator=op, initial_data=initial_data)
    rc = GenericRBReconstructor(rb_N)
    rd.disable_logging()
    return rd, rc, {}


def ei_greedy_pod_greedy(d, samples, ei_samples, N, M):
    # ei-greedy
    sample = tuple(d.parameter_space.sample_uniformly(ei_samples))
    operators = [d.operator, d.very_frozen_operator] if hasattr(d, 'very_frozen_operator') else [d.operator]
    evaluations = d.operator.range.empty()
    for mu in sample:
        U = d.solve(mu)
        for op in operators:
            evaluations.append(op.apply(U, mu=mu))
    dofs, collateral_basis, data = ei_greedy(evaluations, atol=1e-10, max_interpolation_dofs=M)

    # pod-greedy
    extension_algorithm = partial(pod_basis_extension)
    reductor = reduce_frozen if isinstance(d, FrozenDiscretization) else reduce_nonfrozen
    reductor = partial(reductor, collateral_basis=collateral_basis, dofs=dofs)
    initial_data = d.initial_data.as_vector()
    initial_data.scal(1./initial_data.l2_norm()[0])
    greedy_data = greedy(d, reductor, d.parameter_space.sample_uniformly(samples),
                         use_estimator=False, error_norm=lambda U: np.max(d.l2_norm(U)),
                         extension_algorithm=extension_algorithm, max_extensions=N)

    rd, rc, rb = greedy_data['reduced_discretization'], greedy_data['reconstructor'], greedy_data['basis']
    subbasis_reductor = partial(reductor, d, rb=rb)
    return rd, rc, subbasis_reductor


def error_analysis(d, subbasis_reductor, Nmax, cs, count=20, test_set_size=30):
    mus = list(d.parameter_space.sample_randomly(test_set_size))

    def doit(N, M):
        print('N = {}, M = {}: '.format(N, M), end='')
        rrd, rrc, _ = subbasis_reductor(N=N, M=M)
        l2_err_max = -1
        mumax = None
        for mu in mus:
            print('.', end='')
            sys.stdout.flush()
            u = rrd.solve(mu)
            URB = rrc.reconstruct(u)
            U = d.solve(mu)
            l2_err = np.max(d.l2_norm(U - URB))
            l2_err = np.inf if not np.isfinite(l2_err) else l2_err
            if l2_err > l2_err_max:
                l2_err_max = l2_err
                mumax = mu
        print()
        return l2_err_max, mumax
    doit = np.frompyfunc(doit, 2, 2)

    Ns = np.linspace(1, Nmax, count).astype(np.int)
    Ms = np.array([(Ns * c).astype(np.int) for c in cs])
    errs, err_mus = doit(Ns, Ms)
    errs = errs.astype(np.float)
    return Ns, errs


def analyse(d, filename):
    mu = VISUALIZE_MU
    cs = CB_RB_RATIOS
    rd, rc, subbasis_reductor = ei_greedy_pod_greedy(d, RB_SAMPLES, CB_SAMPLES, RB_SIZE, CB_SIZE)

    if mu is not None:
        U = d.solve(mu)
        U_RB = rc.reconstruct(rd.solve(mu))
        d.visualize((U, U_RB), title=filename, legend=('detailed', 'reduced'))

    Ns, errs = error_analysis(d, subbasis_reductor, Nmax=rd.solution_space.dim, cs=cs,
                              count=TEST_DIMS,
                              test_set_size=TEST_SET_SIZE)
    with open(filename, 'wt') as f:
        print('N', *cs, sep='\t', file=f)
        for n, e in zip(Ns, errs.T):
            print(n, *e, sep='\t', file=f)


# MAIN
p = Burgers2DProblem()
d, d_data = discretize_nonlinear_instationary_advection_fv(
    p,
    diameter=DIAMETER,
    nt=NT,
    num_flux='lax_friedrichs',
    lxf_lambda=0.5,
    domain_discretizer=partial(discretize_domain_default, grid_type=RectGrid)
)
d.enable_caching('disk')

d_f = freeze(d, d_data)
d_f.enable_caching('disk')

analyse(d, 'results_non_frozen')
analyse(d_f, 'results_frozen')
